/*
 * ddfschkrep.c
 *
 *  Created on: Jun 4, 2011
 *      Author: asx
 */

#define _XOPEN_SOURCE 500
#define _LARGEFILE64_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>

#include "ddfschkrep.h"

/**
 * order nodes inside the index and remove node with addresses out of ranges
 *
 * @param wrong return the list of blocks that were at wrong place in the index
 * @param verbose if true report all errors
 * @return number of errors
 */
long long int ddfs_fsck_repair_node_order(struct bit_array *wrong, int verbose)
{
    long long int errors=0;
    nodeidx node_idx=0;

    while (node_idx<ddfs.c_node_count)
    {
        unsigned char *node=ddfs.nodes+(node_idx*ddfs.c_node_size);
        unsigned char *hash=node+ddfs.c_addr_size;
        blockaddr addr=ddfs_get_node_addr(node);
        if (addr==0)
        {
            memset(hash, '\0', ddfs.c_hash_size);
        }
        else if (addr<=DDFS_LAST_RESERVED_BLOCK || ddfs.c_block_count<=addr)
        {
            errors++;
            if (verbose) DDFS_LOG(LOG_ERR, "repair_node_order: node %lld (addr=%lld) has an invalid address, delete\n", node_idx, addr);
            node_delete(node_idx);
            continue;
        }
        else
        {
            nodeidx cidx=ddfs_hash2idx(hash);
            if (node_idx<cidx)
            {
                errors++;
                if (verbose) DDFS_LOG(LOG_ERR, "repair_node_order: node %lld (addr=%lld) before it ideal position %lld, fix\n", node_idx, addr, cidx);
                bit_array_set(wrong, addr);
                node_fix(node_idx);
                continue;
            }
            else if (cidx==node_idx)
            {
            }
            else // (cidx<node_idx)
            {
                blockaddr baddr=ddfs_get_node_addr(node-ddfs.c_node_size);
                if (baddr==0 || memcmp(node+(ddfs.c_addr_size-ddfs.c_node_size), hash, ddfs.c_hash_size)>0)
                {
                    errors++;
                    if (verbose) DDFS_LOG(LOG_ERR, "repair_node_order: nodes %lld (addr=%lld) and %lld (addr=%lld) in wrong order, fix.\n", node_idx-1, baddr, node_idx, addr);
                    bit_array_set(wrong, addr);
                    node_fix(node_idx);
                    node_idx=cidx;
                    continue;
                }
            }
        }
        node_idx++;
    }
    return errors;
}

/**
 * verify hash of nodes from list and remove bad nodes
 *
 * @param list the list of blocks to check
 * @param verbose if true report all wrong nodes
 * @return number of errors
 */
long long int ddfs_fsck_check_node_hash(struct bit_array *list, int verbose)
{
    long long int errors=0;
    nodeidx node_idx=0;

    while (node_idx<ddfs.c_node_count)
    {
        unsigned char *node=ddfs.nodes+(node_idx*ddfs.c_node_size);
        unsigned char *hash=node+ddfs.c_addr_size;
        blockaddr addr=ddfs_get_node_addr(node);
        if (addr==0)
        {
        }
        else if (addr<=DDFS_LAST_RESERVED_BLOCK)
        {
            DDFS_LOG(LOG_ERR, "check_node_hash: node %lld (addr=%lld) has a wrong hash, unexpected here, ABORT.\n", node_idx, addr);
            exit(1);
        }
        else if (bit_array_get(list, addr))
        {   // this is a node to check
            int len=ddfs_read_full_block(addr, ddfs.aux_buffer);
            if (len!=ddfs.c_block_size)
            {
                ddfs.error_block_read++;
                errors++;
                DDFS_LOG(LOG_ERR, "cannot read block %lld (%s), delete node %lld from index.\n", addr, len==-1?strerror(errno):"", node_idx);
                // if the block is unreadable, remove it from the index !
                node_delete(node_idx);
                continue;
            }
            // check the hash
            unsigned char bhash[HASH_SIZE];
            ddfs_hash(ddfs.aux_buffer, bhash);
            if (0==memcmp(bhash, hash, ddfs.c_hash_size))
            {
                if (node_idx>0 && 0==memcmp(hash, hash-ddfs.c_node_size, ddfs.c_hash_size))
                {
                    errors++;
                    if (verbose) DDFS_LOG(LOG_ERR, "check_node_hash: node %lld (addr=%lld) is identical to previous one, remove.\n", node_idx, addr);
                    node_delete(node_idx);
                    continue;
                }
            }
            else
            {
                errors++;
                if (verbose) DDFS_LOG(LOG_ERR, "check_node_hash: node %lld (addr=%lld) has a wrong hash, deleted.\n", node_idx, addr);
                node_delete(node_idx);
                continue;
            }
        }
        node_idx++;
    }
    return errors;
}

/**
 * report block addresses that are in use and that are referenced more than once
 *
 * foundblocks2 will contains addresses that are referenced twice
 * or more (different hash) or that are duplicate (same hash) with other
 *
 * @param foundblocks return list of blocks found in the index
 * @param foundblocks2 return list of blocks over-referenced
 */
void ddfs_fsck_node_accounting(struct bit_array *foundblocks, struct bit_array *foundblocks2)
{
    nodeidx node_idx;

    for (node_idx=0; node_idx<ddfs.c_node_count; node_idx++)
    {
        blockaddr addr=ddfs_get_node_addr(ddfs.nodes+(node_idx*ddfs.c_node_size));
        if (addr==0) continue;

        // check if address was already used as reference for another hash
        if (bit_array_get(foundblocks, addr)) bit_array_set(foundblocks2, addr);
        else bit_array_set(foundblocks, addr);

        // check if hash is identical to previous one
        if (addr!=0 && node_idx>0)
        {
            blockaddr paddr=ddfs_get_node_addr(ddfs.nodes+((node_idx-1)*ddfs.c_node_size));
            if (paddr!=0 && 0==memcmp(ddfs.nodes+((node_idx-1)*ddfs.c_node_size+ddfs.c_addr_size), ddfs.nodes+(node_idx*ddfs.c_node_size+ddfs.c_addr_size), ddfs.c_hash_size))
            {
                bit_array_set(foundblocks2, addr);
                bit_array_set(foundblocks2, paddr);
            }
        }
    }
}

/**
 * delete useless node
 *
 * references to unused blocks can be removed
 *
 * @param keep the list of blocks to keep
 * @return number of deleted nodes
 */
long long int ddfs_fsck_remove_useless_node(struct bit_array *keep)
{
    long long int deleted=0;
    nodeidx node_idx=0;

    while (node_idx<ddfs.c_node_count)
    {
        unsigned char *node=ddfs.nodes+(node_idx*ddfs.c_node_size);
        blockaddr addr=ddfs_get_node_addr(node);
        if (addr!=0 && !bit_array_get(keep, addr))
        {
            deleted++;
            node_delete(node_idx);
            continue;

        }
        node_idx++;
    }
    return deleted;
}

long long int ddfs_fsck_te_file_count;
long long int ddfs_fsck_te_addr_count;
long long int ddfs_fsck_te_frag_count;
long long int ddfs_fsck_te_corrupted_count;
int ddfs_fsck_te_verbose;
int ddfs_fsck_te_progress;
int ddfs_fsck_te_update_index_relaxed;
long long int ddfs_fsck_te_fix_file;
FILE *ddfs_fsck_te_corrupted_stream=NULL;

struct bit_array *ddfs_fsck_te_block_found=NULL;
struct bit_array *ddfs_fsck_te_update_index=NULL;

/**
 * function called by nftw to walk through the directory tree
 *
 */
int ddfs_fsck_tree_explore(const char *fpath, const struct stat *sb, int typeflag, struct FTW *ftwbuf)
{
    int res, len;
    unsigned char node[NODE_SIZE];
    unsigned char *hash=node+ddfs.c_addr_size;
    blockaddr addr;
    blockaddr prev_addr=-1;
    uint64_t size;

    if (typeflag!=FTW_F) return 0;
    if (0==strncmp(fpath+ddfs.rdir_len, SPECIAL_DIR, ddfs.special_dir_len))
    {   // skip SPECIAL_DIR
        return 0;
    }

    ddfs_fsck_te_file_count++;
    // fprintf(stderr, "ddfs_fsck_tree_explore file: %lld %s %s\n", ddfs_fsck_te_file_count, fpath, ddfs.rdir);

    FILE *file;
    if (ddfs_fsck_te_fix_file) file=fopen(fpath, "r+");
    else file=fopen(fpath, "r");
    if (file==NULL)
    {
        DDFS_LOG(LOG_ERR, "cannot open: %s (%s)\n", fpath, strerror(errno));
        return 1;
    }

    // be careful, file_header_set_conv return c_file_header_size for empty file
    len=file_header_set_conv(fileno(file), &size);
    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "cannot read header: %s (%s)\n", fpath, strerror(errno));
        fclose(file);
        return 1;
    }
    else if (len!=ddfs.c_file_header_size)
    {
        if (ddfs_fsck_te_verbose) DDFS_LOG(LOG_INFO, "corrupted header, ignore file : %s\n", fpath);
        if (ddfs_fsck_te_corrupted_stream)
        {
            fprintf(ddfs_fsck_te_corrupted_stream, " H   %s\n", fpath);
            DDFS_LOG(LOG_WARNING,  " H   %s\n", fpath);
        }
        ddfs_fsck_te_corrupted_count++;
        fclose(file);
        return 0;
    }

    res=fseek(file, ddfs.c_file_header_size, SEEK_SET);
    if (res==-1)
    {
        DDFS_LOG(LOG_ERR, "cannot seek at first block: %s (%s)\n", fpath, strerror(errno));
        fclose(file);
        return 1;
    }

    if (ddfs_fsck_te_progress)
    {
        fprintf(stdout, "file: %lld\r", ddfs_fsck_te_file_count);
        fflush(stdout);
    }

    long long int block_pos=0;
    int corrupted=0;
    int updated=0;
    int len_mismatch=' ';
    int size_mismatch=' ';
    int cannot_fix_this_file=0;

    len=fread(node, 1, ddfs.c_node_size, file);

    while (len==ddfs.c_node_size)
    {
        addr=ddfs_get_node_addr(node);

        if (prev_addr!=-1 && prev_addr+1!=addr && addr!=0) ddfs_fsck_te_frag_count++;
        if (addr!=0) prev_addr=addr;

        ddfs_fsck_te_addr_count++;

        if (ddfs_fsck_te_update_index || ddfs_fsck_te_update_index_relaxed)
        {
            if (addr<=DDFS_LAST_RESERVED_BLOCK || ddfs.c_block_count<=addr)
            {
            }
            else
            {
                blockaddr baddr;
                nodeidx node_idx=ddfs_search_hash(hash, &baddr);
                if ((node_idx<0 || addr!=baddr) && !(ddfs_fsck_te_update_index && bit_array_get(ddfs_fsck_te_update_index, addr)))
                {
                    // the block has not been already checked
                    if (ddfs_fsck_te_update_index_relaxed)
                    {
                        if (node_add(node)!=0) return 1;
                        if (ddfs_fsck_te_update_index) bit_array_set(ddfs_fsck_te_update_index, addr);
                    }
                    else
                    {

                        DDFS_LOG(LOG_INFO, "calculate hash for block addr=%lld\n", addr);
                        int len=ddfs_read_full_block(addr, ddfs.aux_buffer);
                        if (len==-1)
                        {
                            DDFS_LOG(LOG_ERR, "cannot read block %lld, skip (%s)\n", addr, strerror(errno));
                        }
                        else if (len!=ddfs.c_block_size)
                        {
                            DDFS_LOG(LOG_ERR, "cannot read block %lld, skip\n", addr);
                        }
                        else
                        {
                            unsigned char bhash[HASH_SIZE];
                            ddfs_hash(ddfs.aux_buffer, bhash);
                            if (0==memcmp(bhash, hash, ddfs.c_hash_size))
                            {
                                if (node_add(node)!=0) return 1;
                                bit_array_set(ddfs_fsck_te_update_index, addr);
                            }
                            // don't fix the file here
                        }
                    }
                }
            }
        }

        if (ddfs_fsck_te_fix_file && !cannot_fix_this_file)
        {
            blockaddr baddr=addr;
            if (addr==0)
            {
                if (0!=memcmp(ddfs.null_block_hash, hash, ddfs.c_hash_size) && 0!=memcmp(ddfs.zero_block_hash, hash, ddfs.c_hash_size))
                {
                    addr=1;
                    memcpy(hash, ddfs.null_block_hash, ddfs.c_hash_size);
                }
            }
            else
            {
                ddfs_search_hash(hash, &addr);
            }

            if (baddr!=addr)
            {   // address has changed, must update file
                ddfs_convert_addr(addr, node);
                // set file postition before to write
                res=fseek(file, ddfs.c_file_header_size+block_pos*ddfs.c_node_size, SEEK_SET);
                if (res==-1)
                {
                    DDFS_LOG(LOG_ERR, "cannot fix file: %s, %s\n", fpath, strerror(errno));
                    cannot_fix_this_file=1;
                }
                else
                {
                    len=fwrite(node, 1, ddfs.c_node_size, file);
                    if (len!=ddfs.c_node_size)
                    {
                        if (len==-1)
                        {
                            DDFS_LOG(LOG_ERR, "cannot fix file: %s, %s\n", fpath, strerror(errno));
                            cannot_fix_this_file=1;
                        }
                        else
                        {
                            DDFS_LOG(LOG_ERR, "cannot fix file: %s %d\n", fpath, len);
                            cannot_fix_this_file=1;
                        }
                    }
                    else
                    {
                        updated=1;
                        if (ddfs_fsck_te_verbose) DDFS_LOG(LOG_INFO, "update file: %s #%lld %lld --> %lld \n", fpath, block_pos, baddr, addr);
                    }
                }
                // set file position before next read, required by ANSI C before to switch between read/write
                res=fseek(file, ddfs.c_file_header_size+(block_pos+1)*ddfs.c_node_size, SEEK_SET);
                if (res==-1)
                {
                    DDFS_LOG(LOG_ERR, "cannot reposition file: %s, %s\n", fpath, strerror(errno));
                    fclose(file);
                    return 1;
                }
            }

            if (addr==1 || cannot_fix_this_file) corrupted=1;
        }

        if (ddfs_fsck_te_block_found) bit_array_set(ddfs_fsck_te_block_found, addr); // use the correct address

        block_pos++;
        len=fread(node, 1, ddfs.c_node_size, file);
    }

    if (len==-1)
    {
        DDFS_LOG(LOG_ERR, "cannot read file: %s (%s)\n", fpath, strerror(errno));
        fclose(file);
        return 1;
    }

    if (len!=0)
    {   // incomplete block
        //fprintf(stderr, "incomplete block: %s %d\n", fpath, len);
        len_mismatch='L';
        if (ddfs_fsck_te_fix_file)
        {   // fix the file length
            res=ftruncate(fileno(file), ddfs.c_file_header_size+block_pos*ddfs.c_node_size);
            if (res==-1)
            {
                DDFS_LOG(LOG_ERR, "cannot fix underneath file length: %s (%s)\n", fpath, strerror(errno));
                cannot_fix_this_file=1;
            }
            else
            {
                len_mismatch='l';
                updated=1;
                if (ddfs_fsck_te_verbose) DDFS_LOG(LOG_INFO, "fixed underneath file length mismatch: %s %lld -> %lld\n", fpath,  ddfs.c_file_header_size+block_pos*ddfs.c_node_size+len,  ddfs.c_file_header_size+block_pos*ddfs.c_node_size);
            }
        }
    }

    if (block_pos!=(size+ddfs.c_block_size-1)/ddfs.c_block_size)
    {   // size mismatch
        size_mismatch='S';
        if (ddfs_fsck_te_fix_file)
        {
            size=block_pos*ddfs.c_block_size;
            len=file_header_get_conv(fileno(file), size);
            if (len==-1)
            {
                DDFS_LOG(LOG_ERR, "cannot fix file size mismatch: %s (%s)\n", fpath, strerror(errno));
                cannot_fix_this_file=1;
            }
            else if (len!=ddfs.c_file_header_size)
            {
                DDFS_LOG(LOG_ERR, "cannot fix file size mismatch: %s len=%d/%d\n", fpath, len, ddfs.c_node_size);
                cannot_fix_this_file=1;
            }
            else
            {
                size_mismatch='s';
                updated=1;
                if (ddfs_fsck_te_verbose) DDFS_LOG(LOG_INFO, "fixed file size mismatch: %s %lld -> %lld\n", fpath,  ddfs.c_file_header_size+block_pos*ddfs.c_node_size+len,  ddfs.c_file_header_size+block_pos*ddfs.c_node_size);
            }
        }
    }

    char status[5];
    memset(status, ' ', sizeof(status));
    status[4]='\0';
    int bad=size_mismatch=='S' || len_mismatch=='L' || corrupted || cannot_fix_this_file;
    if (updated && !bad)
    {
        status[0]='F'; // FIXED
        ddfs_fsck_te_fix_file++;
    }
    if (bad) ddfs_fsck_te_corrupted_count++;
    if (corrupted) status[1]='C';
    status[2]=size_mismatch;
    status[3]=len_mismatch;

    if (updated || bad)
    {
        if (ddfs_fsck_te_corrupted_stream) fprintf(ddfs_fsck_te_corrupted_stream, "%s %s\n", status, fpath);
        if (ddfs_fsck_te_fix_file) DDFS_LOG(LOG_WARNING, "%s %s\n", status, fpath);
    }

    fclose(file);
    return 0;
}

/**
 * use valid index to check and fix files, update ba_usedblocks too
 *
 * @return 0 for success
 */
int ddfs_fsck_fix_files_and_usedblocks(int verbose, int progress)
{
    long long int start, end;
    long long int block_count, block_free;
    char filename[FILENAME_MAX];
    int res=0;

    ddfs_fsck_te_file_count=ddfs_fsck_te_addr_count=ddfs_fsck_te_frag_count=ddfs_fsck_te_corrupted_count=0;
    ddfs_fsck_te_verbose=verbose;
    ddfs_fsck_te_progress=progress;

    bit_array_reset(&ddfs.ba_usedblocks, 0x0);
    bit_array_set(&ddfs.ba_usedblocks, 0);  // block ZERO is always used
    bit_array_set(&ddfs.ba_usedblocks, 1);  // block    1 is always used
    ddfs_fsck_te_block_found=&ddfs.ba_usedblocks;
    ddfs_fsck_te_update_index=NULL;
    ddfs_fsck_te_update_index_relaxed=0;
    sprintf(filename, "%s/%s", ddfs.rdir, DDFS_CORRUPTED_LIST);
    xlog_rollout(filename, DDFS_CORRUPTED_BACKUP_COUNT);
    ddfs_fsck_te_corrupted_stream=fopen(filename, "w");
    ddfs_fsck_te_fix_file=1;

    DDFS_LOG(LOG_INFO, "Fix files.\n");
    start=now();
    res=nftw(ddfs.rdir, ddfs_fsck_tree_explore, 10, FTW_MOUNT | FTW_PHYS);
    end=now();
    fclose(ddfs_fsck_te_corrupted_stream);
    if (res)
    {
        res=1;
        DDFS_LOG(LOG_ERR, "nftw was unable to fix all files, correct and check filesystem again !\n");
    }
    else
    {
        DDFS_LOG(LOG_INFO, "Fixed:%lld  Corrupted:%lld  Total:%lld files in %.1fs.\n", ddfs_fsck_te_fix_file-1, ddfs_fsck_te_corrupted_count, ddfs_fsck_te_file_count, (end-start)*1.0/NOW_PER_SEC);
        //
        // Clean up the extra nodes
        //
        long long int deleted=ddfs_fsck_remove_useless_node(&ddfs.ba_usedblocks);
        DDFS_LOG(LOG_INFO, "Deleted %lld useless nodes.\n", deleted);
        bit_array_count(&ddfs.ba_usedblocks, &block_count, &block_free);
        DDFS_LOG(LOG_INFO, "blocks in use: %lld   blocks free: %lld.\n", block_count, block_free);
    }
    return res;
}

int ddfs_fsck(int relaxed, int verbose, int progress)
{
    int res;
    long long int errors;
    long long int start, end;
    long long int s, u;
    struct bit_array ba_found_in_files;
    struct bit_array ba_found_in_nodes;
    struct bit_array ba_suspect;
    struct bit_array ba_rehash;
    struct bit_array ba_backup;

    bit_array_init(&ba_found_in_files, ddfs.c_block_count, 0x0);
    bit_array_init(&ba_found_in_nodes, ddfs.c_block_count, 0x0);
    bit_array_init(&ba_suspect, ddfs.c_block_count, 0x0);
    bit_array_init(&ba_rehash, ddfs.c_block_count, 0x0);
    bit_array_init(&ba_backup, ddfs.c_block_count, 0x0);

    bit_array_reset(&ba_suspect, 0);

    //
    // be sure the index is well ordered
    //
    errors=ddfs_fsck_repair_node_order(&ba_suspect, verbose);
    DDFS_LOG(LOG_INFO, "Repair node order, fixed %lld errors.\n", errors);

    //
    // update Index with verified hashes and addresses from files
    //
    ddfs_fsck_te_file_count=ddfs_fsck_te_addr_count=ddfs_fsck_te_frag_count=ddfs_fsck_te_corrupted_count=0;
    ddfs_fsck_te_verbose=verbose;
    ddfs_fsck_te_progress=progress;
    bit_array_reset(&ba_found_in_files, 0);
    bit_array_reset(&ba_rehash, 0);
    bit_array_set(&ba_found_in_files, 0);  // block ZERO is always used
    bit_array_set(&ba_found_in_files, 1);  // block    1 is always used
    ddfs_fsck_te_block_found=&ba_found_in_files;
    ddfs_fsck_te_update_index=&ba_rehash;
    ddfs_fsck_te_update_index_relaxed=relaxed;

    ddfs_fsck_te_corrupted_stream=NULL;
    ddfs_fsck_te_fix_file=0;

    DDFS_LOG(LOG_INFO, "Update index from files.\n");
    start=now();
    res=nftw(ddfs.rdir, ddfs_fsck_tree_explore, 10, FTW_MOUNT | FTW_PHYS);
    end=now();
    if (res)
    {
        DDFS_LOG(LOG_ERR, "nftw is unable to read all files, correct and check filesystem again !\n");
        res=1;
        goto END;
    }
    bit_array_count(&ba_found_in_files, &s, &u);
    DDFS_LOG(LOG_INFO, "Read %lld files in %.1fs.\n", ddfs_fsck_te_file_count, (end-start)*1.0/NOW_PER_SEC);
    DDFS_LOG(LOG_INFO, "%lld blocks used in files.\n", s);
    bit_array_count(&ba_rehash, &s, &u);
    DDFS_LOG(LOG_INFO, "%lld blocks have been added to index.\n", s);
    // I don't need to reuse ba_rehash later because these blocks are
    // in the index now and are well referenced.

    //
    // Get a list of all blocks referenced by Index
    // ! ba_suspect is already filled with block from the node order check
    bit_array_reset(&ba_found_in_nodes, 0);
    ddfs_fsck_node_accounting(&ba_found_in_nodes, &ba_suspect);

    // I add to ba_suspect all blocks that where recently allocated (since last backup) to
    // be sure these block where "committed"

    DDFS_LOG(LOG_INFO, "ddfs_load_usedblocks\n");
    if (ddfs_load_usedblocks(&ba_backup)==0)
    {
        // ba_suspect=ba_suspect | (~ba_backup ^ ba_found_in_nodes)
        long long int suspect_count1, suspect_count2, u;
        bit_array_count(&ba_suspect, &suspect_count1, &u);
        bit_array_plus_diff(&ba_suspect, &ba_found_in_nodes, &ba_backup);
        bit_array_count(&ba_suspect, &suspect_count2, &u);
        DDFS_LOG(LOG_INFO, "Check also recently added blocks: %lld.\n", suspect_count2-suspect_count1);
    }

    // blocks in ba_suspect are referenced twice or more and need to be verified
    bit_array_set(&ba_found_in_nodes, 0);  // block ZERO and 1
    bit_array_set(&ba_found_in_nodes, 1);  // must be referenced by Index
    bit_array_unset(&ba_suspect, 0);  // block ZERO and 1
    bit_array_unset(&ba_suspect, 1);  // cannot be suspected

    bit_array_count(&ba_found_in_nodes, &s, &u);
    DDFS_LOG(LOG_INFO, "%lld blocks used in nodes.\n", s);
    bit_array_count(&ba_suspect, &s, &u);
    DDFS_LOG(LOG_INFO, "%lld suspect blocks in nodes.\n", s);

    DDFS_LOG(LOG_INFO, "Resolve Index conflicts.\n");
    errors=ddfs_fsck_check_node_hash(&ba_suspect, verbose);
    DDFS_LOG(LOG_INFO, "%lld nodes fixed.\n", errors);

    //
    // update files regarding new index
    //
    res=ddfs_fsck_fix_files_and_usedblocks(verbose, progress);

    //
    // Update used_block list backup
    //
    ddfs_save_usedblocks();

END:
    bit_array_release(&ba_found_in_files);
    bit_array_release(&ba_found_in_nodes);
    bit_array_release(&ba_suspect);
    bit_array_release(&ba_rehash);
    bit_array_release(&ba_backup);

    return res;
}
