/**
 * @testddumbfs.c
 * @author  Alain Spineux <alain.spineux@gmail.com>
 * @version 1.0
 *
 * @section LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details at
 * http://www.gnu.org/copyleft/gpl.html
 *
 * @section DESCRIPTION
 *
 * testddumbfs is tool to test filesystem speed and integrity
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <mhash.h>

#include <fcntl.h>

#define DROP_CACHE "/proc/sys/vm/drop_caches"

struct rbstat
{
    float usage;        // in Mo
    float first_write;  // in Mo/s
    float second_write; // in Mo/s
    float read;         // in Mo/s
} *rbstats;

static struct option long_options[] =
{
       {"help",                 no_argument,       0, 'h'},
       {"verbose",              no_argument,       0, 'v'},

       {"count",                required_argument, 0, 'c'},
       {"block-size",           required_argument, 0, 'B'},
       {"file-size",            required_argument, 0, 'S'},
       {"mask",                 required_argument, 0, 'm'},
       {"operation",            required_argument, 0, 'o'},
       {"seed",                 required_argument, 0, 's'},

       {0, 0, 0, 0}
};

void usage()
{
    printf("Usage: testddumbfs [options] target-directory\n"
            "\n    set of tests for filesystem.\n\nOptions\n"
            "  -h, --help            show this help message and exit\n"
            "  -v, --verbose         be more verbose and display progress\n"
            "  -c, --count=COUNT     repeat the operation COUNT time, default is 4\n"
            "  -B, --block-size=SIZE size of block write/read, default is 128K\n"
            "  -S, --file-size=SIZE  size of file to generate, default 1024M\n"
            "  -m, --mask=MASK       mask to increase compressibility of blocks\n"
            "                        0x00000000 = 100.0%% using gzip -1\n"
            "                        0x60606060 =  77.4%% using gzip -1\n"
            "                        0x60e0e0e0 =  73.9%% using gzip -1 (default)\n"
            "                        0xe0e0e0e0 =  68.1%% using gzip -1\n"
            "  -s, --seed=SEED       random seed, default is 0\n"
            "  -f, --fast            fast random, rewrite only one int32 in each block\n"
            "  -o, --operation=OPS  list of operation to run, must be in \"12R\" order\n"
            "                        1 = first write\n"
            "                        2 = second write (for testing deduplication)\n"
            "                        R = read test\n"
            "                        F = create a single file of specified name and size\n"
            "                        C = compare the specified file, 'seed', 'size'"
            "                            and 'fast' must match parameters used for the 'F'"
            "                            operation at creation time\n"
            "                        default is 12R\n"
    );
}

char *target;
int verbose_flag;
int count=4;
int block_size=128*1024;
long long int file_size=1024*1024*1024;
long int bitmask=0x60e0e0e0;
char *operation="12R";
int rndseed=0;
int fast_random=0;

unsigned long *block_buf=NULL;
unsigned long *block_buf_aux=NULL;
unsigned char *block_hash=NULL;


static unsigned long next=1;

void mysrand(unsigned long seed)
{
    next=seed;
}

inline unsigned long int myrand(void)
{
   next=next * 1103515245 + 12345;
//   next = (214013*next+2531011); // intel
   return next;
}

long long int now()
{
    struct timeval tv;
    struct timezone tz;
    int res=gettimeofday(&tv, &tz);
    if (res==-1) perror("gettimeofday");
    return tv.tv_sec*1000*1000+tv.tv_usec;
}

void free_cache(char *op)
{
    int fd, res;
    fd=open(DROP_CACHE, O_WRONLY);
    if (fd==-1)
    {
        perror(DROP_CACHE);
        return;
    }
    res=write(fd, op, strlen(op));
    if (res!=strlen(op))
    {
        if (res==-1) perror(op);
        fprintf(stderr, "incomplete write to %s.", DROP_CACHE);
    }
    close(fd);
}

int genfile(unsigned long seed, char *filename, unsigned char *hash, int compare)
{
    int err=0;
    MHASH td=MHASH_FAILED;
    if (hash)
    {
        td=mhash_init(MHASH_MD5);
        if (td==MHASH_FAILED) exit(1);
    }

    int fd;
    if (compare) fd=open(filename, O_RDONLY);
    else fd=open(filename, O_WRONLY|O_CREAT, 0644);
    if (fd==-1)
    {
        perror(filename);
        exit(1);
    }
    int n=block_size/sizeof(unsigned long);
    long long int size=0;
    int fast_ixd=0;
    while (size<file_size && !err)
    {
        unsigned long *to=block_buf+n;
        unsigned long *p;

        if (!fast_random || size==0)
        {
            for (p=block_buf; p<to; p++)
            {
                seed=seed * 1103515245 + 12345;
                *p=seed|bitmask;
            }
        }
        if (fast_random && size!=0)
        {
            seed=seed * 1103515245 + 12345;
            *(block_buf+(fast_ixd%n))=seed|bitmask;
            fast_ixd++;
            seed=seed * 1103515245 + 12345;
            *(block_buf+(fast_ixd%n))=seed|bitmask;
            fast_ixd++;
            seed=seed * 1103515245 + 12345;
            *(block_buf+(fast_ixd%n))=seed|bitmask;
            fast_ixd++;
        }
        int len;
        if (compare)
        {
            len=read(fd, block_buf_aux, block_size);
            err=memcmp(block_buf, block_buf_aux, block_size)!=0;
        }
        else len=write(fd, block_buf, block_size);
        if (hash) mhash(td, block_buf, block_size);
        err=(len!=block_size || err);
        size+=block_size;
    }

    fsync(fd);
    close(fd);

    if (hash)
    {
        mhash_deinit(td, hash);
    }
    return err;
}

void print_hash(unsigned char *hash)
{
    int i;
    for (i=0; i<mhash_get_block_size(MHASH_MD5); i++)
    {
        printf("%.2x", hash[i]);
    }
}

int speed_test()
{
    int i, res;
    char filename[1024];
    long long int start, end;


    res=chdir(target);
    if (res==-1)
    {
        perror(target);
        exit(1);
    }

    int err=0;
    int loop=1;

    int stat_count=count;

    if (strchr(operation, '2')!=NULL) loop=2;

//    printf("operation=%s %p %p\n", operation, strchr(operation, '1'), strchr(operation, '2'));

    for (loop=(strchr(operation, '1')!=NULL? 0 : 1); loop<(strchr(operation, '2')!=NULL? 2 : 1); loop++)
    {
        if (loop==0) printf("== first write\n");
        if (loop==1) printf("== second write\n");

        for (i=0; !err && (i<count || count==0); i++)
        {
            mysrand(i+rndseed);
            sprintf(filename, "data%06d-%d.bin", i, loop);
            start=now();

            genfile(i+rndseed, filename, NULL, 0);

            end=now();
            float tp=1000000.0/1048576.0*file_size/(end-start);
            printf("%6d\t%5.2f\t%5.2f\n", i, (end-start)/1000000.0, tp);
            if (stat_count)
            {
                if (loop==0) rbstats[i].first_write=tp;
                if (loop==1) rbstats[i].second_write=tp;
            }
        }
    }

    if (strchr(operation, 'R')!=NULL)
    {
//        free_cache("1\n");
        printf("== read\n");
        for (i=0; !err && (i<count || count==0); i++)
        {
            start=now();
            sprintf(filename, "data%06d-%d.bin", i, 0);
            int fd=open(filename, O_RDONLY, 0644);
            if (fd==-1)
            {
                break;
            }

            int size=0;
            while (size<file_size && !err)
            {
                int len=read(fd, block_buf, block_size);
                err=(len!=block_size);
                size+=block_size;
            }
            close(fd);
            end=now();
            float tp=1000000.0/1048576.0*size/(end-start);
            printf("%6d\t%5.2f\t%5.2f\n", i, (end-start)/1000000.0, tp);
            if (stat_count) rbstats[i].read=tp;
        }
    }

    if (stat_count)
    {
        printf("== stats\n");
        for (i=0; i<count; i++)
        {
            printf("%6d\t%5.2f", i, rbstats[i].first_write);
            if (strchr(operation, '2')!=NULL) printf("\t%5.2f", rbstats[i].second_write);
            if (strchr(operation, 'R')!=NULL) printf("\t%5.2f", rbstats[i].read);
            printf("\n");
        }
    }
    return 0;
}


long long int unitvalue(char *u)
{
    if (u)
    {
        switch (*u)
        {
            case 'k':
            case 'K':
                return 1024;
            case 'm':
            case 'M':
                return 1048576;
            case 'g':
            case 'G':
                return 1073741824;
            case 't':
            case 'T':
                return 1099511627776;
        }
    }
    return 1;
}

int main(int argc, char *argv[])
{
    int c;
    char *unit;

    while (1)
    {
        // getopt_long stores the option index here.
        int option_index=0;
        char *optparse;

        c=getopt_long(argc, argv, "hvc:B:S:m:o:s:f", long_options, &option_index);

        // Detect the end of the options.
        if (c==-1) break;

        switch (c)
        {
            case 'h':
                usage();
                return 0;
                break;

            case 'v':
                verbose_flag=1;
                break;

            case 'f':
                fast_random=1;
                break;

            case 'c':
                count=atoi(optarg);
                break;

            case 'B':
                block_size=strtol(optarg, &unit, 10);
                block_size*=unitvalue(unit);
                break;

            case 'S':
                file_size=strtol(optarg, &unit, 10);
                file_size*=unitvalue(unit);
                break;

            case 'm':
                bitmask=strtoll(optarg, &optparse, 0);
                break;

            case 'o':
                operation=strdup(optarg);
                break;

            case 's':
                rndseed=atoi(optarg);
                break;

            default:
                abort();
        }
    }

    if (optind>=argc)
    {
        fprintf(stderr, "Target directory is missing !\n");
        return 1;
    }

    target=argv[optind];

    rbstats=malloc(sizeof(struct rbstat)*count);

    block_buf=malloc(block_size);
    block_buf_aux=malloc(block_size);
    block_hash=malloc(mhash_get_block_size(MHASH_MD5));
    if (block_buf==NULL || block_buf_aux==NULL || block_hash==NULL)
    {
        perror("allocating buffer");
        exit(1);
    }

    int ret=0;

    if (operation[0]=='1' || operation[0]=='2' || operation[0]=='R')
    {
        fprintf(stderr, "seed=%d bitmaks=0x%lx operation=%s\n", rndseed, bitmask, operation);
        ret=speed_test();
    }
    else if (operation[0]=='F' || operation[0]=='C')
    {
        ret=genfile(rndseed, target, NULL, operation[0]=='C');
    }

    return ret;
}
